#ifndef FILEIO_H
#define FILEIO_H

#include <vector>
#include <string.h>
#include <stdio.h>
#include <map>
#include <iomanip>
#include "Splines.h"

/*	Since a vertex's position in a curve is listed before we know
 *	its coordinates (its listed in the permutation line), we need 
 *	to store this information for when we read this info.
 */
struct vertexInfo
{
	int curveNum;	
	int posInCurve;
};

class FileIO
{
	char *writefile;	///< The name of the file that we are writing to
	char *readfile;		///< The name of the file that we are reading from		

public:
	/*
	 *	Default constructor
	 */
	FileIO(){
		writefile=new char();
		readfile=new char();
		strcpy(writefile, "");
		strcpy(readfile, "");
	}

	/*
	 *	Writes the control points to a file so they can later be used in ilevel
	 *	@param spline The spline data for every single level
	 */
	void writeFile(std::vector<Splines *> &spline)
	{
		if(!strcmp(writefile, ""))	// if we do not have a filename to write to, get one
		{
			printf("Output filename: ");
			scanf("%s", writefile);
		}
		// open up the file with write permissions (if this file already exists, this destroys it)
		FILE *fileptr = fopen(writefile, "w");	

		int numNodes=0;				// keeps track of the number of nodes on a level
		for(int i =0; i<spline.size(); i++)	// go through all of the levels
		{
			// go through each curve on level i
			for(int j = 0; j<spline[i]->levelCurves.size(); j++)
			{
				fprintf(fileptr, "( ");	// start a cycle
				// go through each node in curve j
				for(int k=0; k<spline[i]->levelCurves[j].controlPoints.size(); k++)
				{
					// print the order of the nodes in the cycle
					fprintf(fileptr, "%i ", numNodes);
					numNodes++;
				}
				fprintf(fileptr, ") " );	// close the curve
				                           //july 9, 2002 added space after parenthesis to comply with combo input parameters-Wendy Hubbard
			}
			fprintf(fileptr, "\n");		// we've finished writing the permutations
			numNodes=0;
			
			// go through each curve on level i
			for(j = 0; j<spline[i]->levelCurves.size(); j++)
			{
				// go through all of the nodes on level j
				for(int k=0; k<spline[i]->levelCurves[j].controlPoints.size(); k++)	
				{
					// Print the x, y, and z coordinates to the file
					fprintf(fileptr, "%i %0.4f %0.4f %0.4f\n", numNodes, 
						spline[i]->levelCurves[j].controlPoints[k].pos[0], 
						spline[i]->levelCurves[j].controlPoints[k].pos[1], 
						spline[i]->levelCurves[j].controlPoints[k].pos[2]);
					numNodes++;
				}
			}
			fprintf(fileptr,"\n");
			numNodes=0;
		}
		fclose(fileptr);	// close the file
	}
	
	/*
	 *	Read in control points from a file
	 *	@param spline The spline data for every single level
	 */
	void readFile(std::vector<Splines *> &spline)
	{
		if(!strcmp(readfile, ""))	// if we do not have a filename to read from, get one
		{
			printf("Input filename: ");
			scanf("%s", readfile);
		}

		FILE *fileptr;
		// open the file as read only if it exists
		if ( (fileptr = fopen(readfile, "r")) == NULL)
		{
			printf("Can't open %s\n", readfile);
			return;
		}

		spline.erase(spline.begin(), spline.end());	// Clear out any previous control points
		char ch;
		int number;			// The number of the node
		int curve=0;			// Curve that we are creating
		int level=0;			// Level that we are creating
		int position=0;			// Position of the vertex in the curve
		vertexInfo point;
		float vertices[3];		// Holds three vertices temporarily for the nodes
		
		// Provides a mapping between the vertex number and its associated data
		std::map <int, vertexInfo> curveInfo;  
		
		while(!feof(fileptr))		// feof() is the end of the file (stdio.h)
		{
			if ( (ch=getc(fileptr)) == '(' )// This line of the file lists the cycles
			{
				while (ch != '\n')	// Read all cycles in
				{
					ch=getc(fileptr);	// get a character from the file
					if(ch == ')' )		// This cycle is done
					{
						curve++;	// we're on a new curve
						// the next point will be at position 0 in the new curve
						position=0;
					}
					else if ( isdigit(ch) )	// We read in a number
					{
						// put the number back into the file because it could be
						// more than one digit
						ungetc(ch, fileptr);
						int pointnum;	// the vertex number to be read in
						fscanf(fileptr, "%d", &pointnum);
						// save the number of the curve that this vertex belongs to
						point.curveNum=curve;
						// save the position of this vertex in the curve
						point.posInCurve=position;
					// create a mapping between this vertex and its curve and position
						curveInfo[pointnum] = point;
						position++;
					}
				}
				// we've read in the permutations for this level
				level++;
				curve=0;
			}

			// deal with the lines that contain the vertex info, structured as:
			// vertex number	x	y	z
			else if (ch != '\n' && ch != EOF  && ch != ' ')
			{
				// put the number back into the file because it could be
				// more than one digit
				ungetc(ch, fileptr);
				// grab the x, y, z coordinates
				fscanf(fileptr, "%d %f %f %f", &number, 
						&vertices[0], &vertices[1], &vertices[2]);
				// retrieve the curve location for this vertex number
				point = curveInfo[number];
				if(spline.size() <= (level-1))	// Nodes have not been entered on this level
				{
					// create a new level and store it
					spline.push_back(new Splines());
					// the constructor Splines() creates levels as being level 0,
					// so fix this
					spline[level-1]->level=level-1;
				}
				// if we have not entered any points for this specific curve
				if(spline[level-1]->levelCurves.size() <= point.curveNum)
				{
					Nodes newNode; 
					newNode.closure=0;
					// make this level of spline large enough to hold all curves
					spline[level-1]->levelCurves.resize(point.curveNum+1);
					spline[level-1]->levelCurves[point.curveNum]=newNode;
				}
				// if the controlPoints vector is too small, increase its size
				if(spline[level-1]->levelCurves[point.curveNum].controlPoints.size() <= point.posInCurve)
					spline[level-1]->levelCurves[point.curveNum].controlPoints.resize(point.posInCurve+1);
				// We're not editing the z-coord in the editor, so just replace
				// it with the level we are working with
				Vec3 verts(vertices[0], vertices[1], level-1);
				// store the vertices
				spline[level-1]->levelCurves[point.curveNum].controlPoints[point.posInCurve] = verts;
				ch=getc(fileptr);	// clear the end of line delimiter
			}
		}
		fclose(fileptr);	// close the file
	}

	/*
	 *	Sets the name of the file that we want to write out to.
	 *	@param filename The name of the file that we'll write to.
	 */
	void setOutputName(char *filename){strcpy(writefile, filename);}

	/*
	 *	Sets the name of the file that we want to read from.
	 *	@param filename The name of the file that we'll read from.
	 */
	void setInputName(char *filename){strcpy(readfile, filename);}

	/*	Returns the name of the file that we are reading from.
	 */
	char* getInputName(){return readfile;}
};	// FileIO



#endif //FILEIO_H
